@@ -1,30 +1,26 @@
 import pathlib
-import sys
-from typing import Optional
+from typing import List, Optional
 
 import pytest
 from pytest import param
 
 from .. import transform as tf
 from ..parse import parse_source_code
-from .conftest import get_grammar
-
-try:
-    import apischema
-except ImportError:
-    # apischema is optional for serialization testing
-    apischema = None
-
-
-# TODO: apischema serialization is recursing infinitely on 3.9 and 3.10;
-# need to dig into details and report it (first test that fails is ARRAY-related)
-APISCHEMA_SKIP = sys.version_info[:2] >= (3, 9)
+from . import conftest
 
 TEST_PATH = pathlib.Path(__file__).parent
 
 
 def roundtrip_rule(rule_name: str, value: str, expected: Optional[str] = None):
-    parser = get_grammar(start=rule_name)
+    """
+    Round-trip a blark grammar rule for testing purposes.
+
+    1. Parse and transform the source code in ``value`` into dataclasses
+    2. Ensure that the dataclass source code representation is identical
+       to the input
+    3. Run serialization/deserialization checks (if enabled)
+    """
+    parser = conftest.get_grammar(start=rule_name)
     transformed = parse_source_code(value, parser=parser)
     print("\n\nTransformed:")
     print(repr(transformed))
@@ -32,15 +28,12 @@ def roundtrip_rule(rule_name: str, value: str, expected: Optional[str] = None):
     print(transformed)
     if expected is None:
         expected = value
-    assert str(transformed) == expected
+    assert str(transformed) == expected, \
+        "Transformed object does not produce identical source code"
 
-    if apischema is not None and not APISCHEMA_SKIP:
-        serialized = apischema.serialize(transformed)
-        print("serialized", serialized)
-        deserialized = apischema.deserialize(type(transformed), serialized)
-        print("deserialized", deserialized)
-        assert str(transformed) == str(deserialized)
-        # assert transformed == deserialized
+    conftest.check_serialization(
+        transformed, deserialize=True, require_same_source=True
+    )
     return transformed
 
 
@@ -63,6 +56,8 @@ def test_check_unhandled_rules(grammar):
         "hours",
         "seconds",
         "milliseconds",
+        "microseconds",
+        "nanoseconds",
         "minutes",
 
         # handled as aliases
@@ -71,7 +66,6 @@ def test_check_unhandled_rules(grammar):
         # handled as tree
         "global_var_list",
         "var_body",
-
     }
 
     todo_rules = set()
@@ -100,6 +94,7 @@ def test_check_unhandled_rules(grammar):
         param("real_literal", "-12.0", tf.Real(value="-12.0")),
         param("real_literal", "12.0", tf.Real(value="12.0")),
         param("real_literal", "12.0e5", tf.Real(value="12.0e5")),
+        param("bit_string_literal", "1234", tf.BitString(type_name=None, value="1234")),
         param("bit_string_literal", "WORD#1234", tf.BitString(type_name="WORD", value="1234")),
         param("bit_string_literal", "WORD#2#0101", tf.BinaryBitString(type_name="WORD", value="0101")),  # noqa: E501
         param("bit_string_literal", "WORD#8#777", tf.OctalBitString(type_name="WORD", value="777")),  # noqa: E501
@@ -118,6 +113,20 @@ def test_check_unhandled_rules(grammar):
         param("duration", "T#1D1H1M1S1MS", tf.Duration(days="1", hours="1", minutes="1", seconds="1", milliseconds="1")),  # noqa: E501
         param("duration", "TIME#1H1M1S1MS", tf.Duration(hours="1", minutes="1", seconds="1", milliseconds="1")),  # noqa: E501
         param("time_of_day", "TIME_OF_DAY#1:1:1.2", tf.TimeOfDay(hour="1", minute="1", second="1.2")),  # noqa: E501
+        param("lduration", "LTIME#-1D", tf.Lduration(days="1", negative=True)),
+        param("lduration", "LTIME#1D", tf.Lduration(days="1")),
+        param("lduration", "LTIME#10S", tf.Lduration(seconds="10")),
+        param("lduration", "LTIME#1H", tf.Lduration(hours="1")),
+        param("lduration", "LTIME#1M", tf.Lduration(minutes="1")),
+        param("lduration", "LTIME#10MS", tf.Lduration(milliseconds="10")),
+        param("lduration", "LTIME#1.1D", tf.Lduration(days="1.1")),
+        param("lduration", "LTIME#10.1S", tf.Lduration(seconds="10.1")),
+        param("lduration", "LTIME#1.1H", tf.Lduration(hours="1.1")),
+        param("lduration", "LTIME#1.1M", tf.Lduration(minutes="1.1")),
+        param("lduration", "LTIME#10.1MS", tf.Lduration(milliseconds="10.1")),
+        param("lduration", "LT#1D1H1M1S1MS", tf.Lduration(days="1", hours="1", minutes="1", seconds="1", milliseconds="1")),  # noqa: E501
+        param("lduration", "LTIME#1H1M1S1MS", tf.Lduration(hours="1", minutes="1", seconds="1", milliseconds="1")),  # noqa: E501
+        param("ltime_of_day", "LTIME_OF_DAY#1:1:1.2", tf.LtimeOfDay(hour="1", minute="1", second="1.2")),  # noqa: E501
     ],
 )
 def test_literal(name, value, expected):
@@ -160,6 +169,10 @@ def test_literal(name, value, expected):
         param("time_of_day", "TIME_OF_DAY#1:1:1.2"),
         param("date", "DATE#1970-1-1"),
         param("date_and_time", "DT#1970-1-1-1:2:30.3"),
+        param("date_and_time", "DT#1970-1-1-0:0:0"),
+        param("date_and_time", "DT#1970-1-1-0:0"),
+        param("ldate", "LDATE#1970-1-1"),
+        param("ldate_and_time", "LDT#1970-1-1-1:2:30.300123456"),
         param("single_byte_string_spec", "STRING[1]"),
         param("single_byte_string_spec", "STRING(1)"),
         param("single_byte_string_spec", "STRING(1) := 'abc'"),
@@ -209,6 +222,9 @@ def test_bool_literal_roundtrip(name, value, expected):
         param("multi_element_variable", "a^.b[1, 2]"),
         param("multi_element_variable", "a[1, 2, 3, 4]^.b[1, 2]"),
         param("multi_element_variable", "Abc123[1]^._defGhi"),
+        param("expression", "-1"),
+        param("expression", "+1"),
+        param("expression", "1"),
         param("expression", "1 + 1"),
         param("expression", "1 / 2"),
         param("expression", "3 * 4"),
@@ -229,7 +245,10 @@ def test_bool_literal_roundtrip(name, value, expected):
         param("simple_type_declaration", "TypeName : INT := 5 + 1 * (2)"),
         param("simple_type_declaration", "TypeName : REFERENCE TO INT"),
         param("simple_type_declaration", "TypeName : POINTER TO INT"),
+        param("simple_type_declaration", "TypeName : POINTER TO POINTER TO INT"),
+        param("simple_type_declaration", "TypeName : REFERENCE TO POINTER TO INT"),
         param("simple_type_declaration", "TypeName EXTENDS a.b : POINTER TO INT"),
+        param("subrange_specification", "TypeName"),  # aliased and not usually hit
         param("subrange_type_declaration", "TypeName : INT (1..2)"),
         param("subrange_type_declaration", "TypeName : INT (*) := 1"),
         param("enumerated_type_declaration", "TypeName : TypeName := Value"),
@@ -240,6 +259,7 @@ def test_bool_literal_roundtrip(name, value, expected):
         param("array_type_declaration", "TypeName : ARRAY [1..2] OF INT := [1, 2]"),
         param("array_type_declaration", "TypeName : ARRAY [1..2, 3..4] OF INT := [2(3), 3(4)]"),
         param("array_type_declaration", "TypeName : ARRAY [1..2, 3..4] OF Tc.SomeType"),
+        param("array_type_declaration", "TypeName : ARRAY [1..2, 3..4] OF Tc.SomeType(someInput := 3)"),  # noqa: E501
         param("structure_type_declaration", "TypeName :\nSTRUCT\nEND_STRUCT"),
         param("structure_type_declaration", "TypeName EXTENDS Other.Type :\nSTRUCT\nEND_STRUCT"),
         param("structure_type_declaration", "TypeName : POINTER TO\nSTRUCT\nEND_STRUCT"),
@@ -463,6 +483,35 @@ def test_var_access_roundtrip(rule_name, value):
     roundtrip_rule(rule_name, value)
 
 
+@pytest.mark.parametrize(
+    "rule_name, value",
+    [
+        param(
+            "static_var_declarations",
+            tf.multiline_code_block(
+                """
+                VAR_STAT
+                    iValue : INT := 1;
+                END_VAR
+                """
+            ),
+        ),
+        param(
+            "static_var_declarations",
+            tf.multiline_code_block(
+                """
+                VAR_STAT CONSTANT
+                    iValue : INT := 1;
+                END_VAR
+                """
+            ),
+        ),
+    ],
+)
+def test_var_stat_roundtrip(rule_name, value):
+    roundtrip_rule(rule_name, value)
+
+
 @pytest.mark.parametrize(
     "rule_name, value",
     [
@@ -472,6 +521,12 @@ def test_var_access_roundtrip(rule_name, value):
             END_VAR
             """
         )),
+        param("global_var_declarations", tf.multiline_code_block(
+            """
+            VAR_GLOBAL INTERNAL
+            END_VAR
+            """
+        )),
         param("global_var_declarations", tf.multiline_code_block(
             """
             VAR_GLOBAL CONSTANT
@@ -503,7 +558,75 @@ def test_var_access_roundtrip(rule_name, value):
         )),
     ],
 )
-def test_global_roundtrip(rule_name, value):
+def test_global_roundtrip(rule_name: str, value: str):
+    gvl = roundtrip_rule(rule_name, value)
+
+    print(gvl.attribute_pragmas)
+
+
+@pytest.mark.parametrize(
+    "rule_name, value, pragmas",
+    [
+        param(
+            "global_var_declarations",
+            tf.multiline_code_block(
+                """
+                VAR_GLOBAL CONSTANT PERSISTENT
+                    iValue : INT := 5;
+                END_VAR
+                """,
+            ),
+            [],
+            id="no_attrs",
+        ),
+        param(
+            "global_var_declarations",
+            tf.multiline_code_block(
+                """
+                {attribute abc}
+                VAR_GLOBAL CONSTANT PERSISTENT
+                    iValue : INT := 5;
+                END_VAR
+                """,
+            ),
+            ["abc"],
+            id="one_attr",
+        ),
+        param(
+            "global_var_declarations",
+            tf.multiline_code_block(
+                """
+                // Line one
+                {attribute abc}
+                // Line two
+                {attribute def}
+                VAR_GLOBAL CONSTANT PERSISTENT
+                    iValue : INT := 5;
+                END_VAR
+                """,
+            ),
+            ["abc", "def"],
+            id="two_attrs",
+        ),
+    ],
+)
+def test_global_attr_pragmas(rule_name: str, value: str, pragmas: List[str]):
+    gvl = roundtrip_rule(rule_name, value)
+    assert gvl.attribute_pragmas == pragmas
+
+
+@pytest.mark.parametrize(
+    "rule_name, value",
+    [
+        param("non_generic_type_name", "POINTER TO FBName"),
+        param("non_generic_type_name", "POINTER TO POINTER TO FBName"),
+        param("non_generic_type_name", "FBName"),
+        param("non_generic_type_name", "POINTER TO Package.FBName"),
+        param("non_generic_type_name", "POINTER TO POINTER TO Package.FBName"),
+        param("non_generic_type_name", "Package.FBName"),
+    ],
+)
+def test_type_name_roundtrip(rule_name, value):
     roundtrip_rule(rule_name, value)
 
 
@@ -534,6 +657,36 @@ def test_global_roundtrip(rule_name, value):
             END_FUNCTION_BLOCK
             """
         )),
+        param("function_block_type_declaration", tf.multiline_code_block(
+            """
+            FUNCTION_BLOCK PRIVATE fbName EXTENDS OtherFbName
+            END_FUNCTION_BLOCK
+            """
+        )),
+        param("function_block_type_declaration", tf.multiline_code_block(
+            """
+            FUNCTION_BLOCK PUBLIC fbName EXTENDS OtherFbName
+            END_FUNCTION_BLOCK
+            """
+        )),
+        param("function_block_type_declaration", tf.multiline_code_block(
+            """
+            FUNCTION_BLOCK INTERNAL fbName EXTENDS OtherFbName
+            END_FUNCTION_BLOCK
+            """
+        )),
+        param("function_block_type_declaration", tf.multiline_code_block(
+            """
+            FUNCTION_BLOCK PROTECTED fbName EXTENDS OtherFbName
+            END_FUNCTION_BLOCK
+            """
+        )),
+        param("function_block_type_declaration", tf.multiline_code_block(
+            """
+            FUNCTION_BLOCK FINAL fbName EXTENDS OtherFbName
+            END_FUNCTION_BLOCK
+            """
+        )),
         param("function_block_type_declaration", tf.multiline_code_block(
             """
             FUNCTION_BLOCK fbName
@@ -717,12 +870,52 @@ def test_global_roundtrip(rule_name, value):
             END_PROPERTY
             """
         )),
+        param("fb_decl", tf.multiline_code_block(
+            """
+            fb1, fb2 : TypeName
+            """
+        )),
+        param("fb_decl", tf.multiline_code_block(
+            """
+            fb1, fb2 : TypeName := (1, 2)
+            """
+        )),
     ],
 )
 def test_fb_roundtrip(rule_name, value):
     roundtrip_rule(rule_name, value)
 
 
+@pytest.mark.parametrize(
+    "rule_name, value",
+    [
+        param(
+            "action",
+            tf.multiline_code_block(
+                """
+                ACTION actActionName:
+                END_ACTION
+                """
+            ),
+            id="empty_action",
+        ),
+        param(
+            "action",
+            tf.multiline_code_block(
+                """
+                ACTION actActionName:
+                    iValue := iValue + 1;
+                END_ACTION
+                """
+            ),
+            id="simple_action",
+        ),
+    ],
+)
+def test_action_roundtrip(rule_name, value):
+    roundtrip_rule(rule_name, value)
+
+
 @pytest.mark.parametrize(
     "rule_name, value",
     [
@@ -754,6 +947,34 @@ def test_fb_roundtrip(rule_name, value):
             END_IF
             """
         )),
+        param("if_statement", tf.multiline_code_block(
+            """
+            IF callable() THEN
+                y();
+            END_IF
+            """
+        )),
+        param("if_statement", tf.multiline_code_block(
+            """
+            IF a() AND b(in := 5) <> 0 AND c(in1 := expr, in2 := 'test') THEN
+                y();
+            END_IF
+            """
+        )),
+        param("if_statement", tf.multiline_code_block(
+            """
+            IF 1 AND_THEN 1 THEN
+                y();
+            END_IF
+            """
+        )),
+        param("if_statement", tf.multiline_code_block(
+            """
+            IF 0 OR_ELSE 1 THEN
+                y();
+            END_IF
+            """
+        )),
         param("case_statement", tf.multiline_code_block(
             """
             CASE expr OF
@@ -775,6 +996,25 @@ def test_fb_roundtrip(rule_name, value):
             END_CASE
             """
         )),
+        param("case_statement", tf.multiline_code_block(
+            """
+            CASE expr OF
+            BYTE#9..BYTE#10, BYTE#13, BYTE#28:
+                OneToTen := OneToTen + 1;
+            EnumValue:
+            END_CASE
+            """
+        )),
+        param("case_statement", tf.multiline_code_block(
+            """
+            CASE expr OF
+            TRUE:
+                abc();
+            FALSE:
+                def();
+            END_CASE
+            """
+        )),
         param("while_statement", tf.multiline_code_block(
             """
             WHILE expr
@@ -843,6 +1083,17 @@ def test_statement_roundtrip(rule_name, value):
             """),
             id="int_with_input",
         ),
+        param("function_declaration", tf.multiline_code_block(
+            """
+            FUNCTION INTERNAL FuncName : INT
+                VAR_INPUT
+                    iValue : INT := 0;
+                END_VAR
+                FuncName := iValue;
+            END_FUNCTION
+            """),
+            id="int_with_input",
+        ),
         param("function_declaration", tf.multiline_code_block(
             """
             FUNCTION FuncName : POINTER TO INT
@@ -885,6 +1136,16 @@ def test_statement_roundtrip(rule_name, value):
             """),
             id="no_return_type",
         ),
+        param("function_declaration", tf.multiline_code_block(
+            """
+            FUNCTION FuncName : BOOL
+                VAR
+                    _c_epoch : dot.dateTime_t := (dateTime := DT#1970-1-1-0:0:0, uSec := 0);
+                END_VAR
+            END_FUNCTION
+            """),
+            id="default_with_datetime",
+        ),
         param("function_declaration", tf.multiline_code_block(
             """
             FUNCTION FuncName : Tc2_System.T_MaxString
@@ -1040,6 +1301,18 @@ def test_incomplete_located_var_decls(rule_name, value):
             END_TYPE
             """
         )),
+        param(
+            "data_type_declaration",
+            tf.multiline_code_block(
+                """
+            TYPE TypeName :
+                UNION
+                END_UNION
+            END_TYPE
+            """
+            ),
+            id="empty_union",   # TODO: this may not be grammatical
+        ),
         param("data_type_declaration", tf.multiline_code_block(
             """
             TYPE TypeName :
@@ -1050,12 +1323,54 @@ def test_incomplete_located_var_decls(rule_name, value):
             END_TYPE
             """
         )),
+        param("data_type_declaration", tf.multiline_code_block(
+            """
+            TYPE TypeName :
+                UNION
+                    pt_value : POINTER TO INT;
+                    pt_SomethingElse : POINTER TO class_SomethingCool;
+                END_UNION
+            END_TYPE
+            """
+        )),
+        param("data_type_declaration", tf.multiline_code_block(
+            """
+            TYPE TypeName :
+                STRUCT
+                    object : class_InitializeWithFBInit(input := one);
+                    interval : INT;
+                END_STRUCT
+            END_TYPE
+            """
+        )),
+        param("data_type_declaration", tf.multiline_code_block(
+            """
+            TYPE INTERNAL EnumTypeName : (ENUM_VAL_1 := 0, ENUM_VAL_2);
+            END_TYPE
+            """
+        )),
     ],
 )
 def test_data_type_declaration(rule_name, value):
     roundtrip_rule(rule_name, value)
 
 
+@pytest.mark.parametrize(
+    "rule_name, value",
+    [
+        param("structure_element_initialization", "1"),
+        param("structure_element_initialization", "name := 1"),
+        param("structure_element_initialization", "name := 1 + 2"),
+        param("structure_element_initialization", "name := GVL.Constant"),
+        param("structure_element_initialization", "name := [1, 2, 3]"),
+        # hmm - aliased by array_initialization?
+        # param("structure_element_initialization", "name := (a:=1, b:=2, c:=3)"),
+    ],
+)
+def test_miscellaneous(rule_name, value):
+    roundtrip_rule(rule_name, value)
+
+
 @pytest.mark.parametrize(
     "value, init, base_type, full_type",
     [
@@ -1078,7 +1393,7 @@ def test_data_type_declaration(rule_name, value):
         ),
         param(
             "fValue : FB_Test(A := 1, B := 2, C => 3);",
-            tf.FunctionBlockInvocation,
+            tf.FunctionCall,
             "FB_Test",
             "FB_Test",
         ),
@@ -1096,7 +1411,7 @@ def test_data_type_declaration(rule_name, value):
     ]
 )
 def test_global_types(value, init, base_type, full_type):
-    parser = get_grammar(start="global_var_decl")
+    parser = conftest.get_grammar(start="global_var_decl")
     transformed = parse_source_code(value, parser=parser)
     assert isinstance(transformed, tf.GlobalVariableDeclaration)
     assert transformed.variables == ["fValue"]
@@ -1106,3 +1421,80 @@ def test_global_types(value, init, base_type, full_type):
     assert transformed.spec.variables
     assert transformed.base_type_name == base_type
     assert transformed.full_type_name == full_type
+
+
+@pytest.mark.parametrize(
+    "code, comments, pragmas",
+    [
+        param(
+            """
+            FUNCTION_BLOCK test
+            END_FUNCTION_BLOCK
+            """,
+            [],
+            [],
+            id="no_comments",
+        ),
+        param(
+            """
+            (* Comment *)
+            FUNCTION_BLOCK test
+            END_FUNCTION_BLOCK
+            """,
+            ["(* Comment *)"],
+            [],
+            id="multiline_comment",
+        ),
+        param(
+            """
+            // Comment
+            FUNCTION_BLOCK test
+            END_FUNCTION_BLOCK
+            """,
+            ["// Comment"],
+            [],
+            id="single_line_comment",
+        ),
+        param(
+            """
+            // Comment
+            (* Comment *)
+            FUNCTION_BLOCK test
+            END_FUNCTION_BLOCK
+            """,
+            ["// Comment", "(* Comment *)"],
+            [],
+            id="both_comments",
+        ),
+        param(
+            """
+            // Comment
+            (* Comment *)
+            {pragma}
+            FUNCTION_BLOCK test
+            END_FUNCTION_BLOCK
+            """,
+            ["// Comment", "(* Comment *)"],
+            ["{pragma}"],
+            id="both_comments_and_pragma1",
+        ),
+        param(
+            """
+            // Comment
+            {pragma}
+            (* Comment *)
+            FUNCTION_BLOCK test
+            END_FUNCTION_BLOCK
+            """,
+            ["// Comment", "(* Comment *)"],
+            ["{pragma}"],
+            id="both_comments_and_pragma2",
+        ),
+    ]
+)
+def test_meta(code: str, comments: List[str], pragmas: List[str]):
+    transformed = parse_source_code(code)
+    meta = transformed.items[0].meta
+    found_comments, found_pragmas = meta.get_comments_and_pragmas()
+    assert [str(comment) for comment in found_comments] == comments
+    assert [str(pragma) for pragma in found_pragmas] == pragmas
