@@ -9,10 +9,12 @@ import typing
 from dataclasses import dataclass, fields, is_dataclass
 from enum import Enum
 from typing import (Any, Callable, ClassVar, Dict, Generator, List, Optional,
-                    Set, Tuple, Type, TypeVar, Union)
+                    Tuple, Type, TypeVar, Union)
 
 import lark
 
+from .util import AnyPath
+
 T = TypeVar("T")
 
 try:
@@ -33,7 +35,6 @@ _class_handlers = {}
 _comment_consumers = []
 
 INDENT = "    "  # TODO: make it configurable
-StringOrToken = Union[str, lark.Token]
 
 
 def multiline_code_block(block: str) -> str:
@@ -79,7 +80,7 @@ def _add_comments_to_return_value(func):
         )
 
     if getattr(func, "_comment_wrapped", None):
-        return func
+        return func  # pragma: no cover
 
     wrapped._comment_wrapped = True
     return wrapped
@@ -104,7 +105,9 @@ def _rule_handler(
         for rule in rules:
             handler = _rule_to_class.get(rule, None)
             if handler is not None:
-                raise ValueError(f"Handler already specified for: {rule} ({handler})")
+                raise ValueError(
+                    f"Handler already specified for: {rule} ({handler})"
+                )  # pragma: no cover
 
             _rule_to_class[rule] = cls
             _class_handlers[rule] = cls.from_lark
@@ -151,7 +154,7 @@ class Meta:
             start_pos=getattr(lark_meta, "start_pos", None),
         )
 
-    def get_comments_and_pragmas(self):
+    def get_comments_and_pragmas(self) -> Tuple[List[lark.Token], List[lark.Token]]:
         """
         Split the contained comments into comments/pragmas.
 
@@ -176,6 +179,18 @@ class Meta:
 
         return comments, pragmas
 
+    @property
+    def attribute_pragmas(self) -> List[str]:
+        """Get {attribute ...} pragmas associated with this code block."""
+
+        _, pragmas = self.get_comments_and_pragmas()
+        attributes = []
+        for pragma in pragmas:
+            # TODO: better pragma parsing; it's its own grammar
+            if pragma.startswith("{attribute "):  # }
+                attributes.append(pragma.split(" ")[1].strip(" }'"))
+        return attributes
+
 
 def meta_field():
     """Create the Meta field for the dataclass magic."""
@@ -199,18 +214,26 @@ class _FlagHelper:
 
 
 @_rule_handler("variable_attributes")
-class VariableAttributes(_FlagHelper, enum.Flag):
+class VariableAttributes(_FlagHelper, enum.IntFlag):
     constant = 0b0000_0001
     retain = 0b0000_0010
     non_retain = 0b0000_0100
     persistent = 0b0000_1000
 
 
+@_rule_handler("global_variable_attributes")
+class GlobalVariableAttributes(_FlagHelper, enum.IntFlag):
+    constant = 0b0000_0001
+    retain = 0b0000_0010
+    non_retain = 0b0000_0100
+    persistent = 0b0000_1000
+    internal = 0b0001_0000
+
+
 @_rule_handler(
-    "method_access",
-    "property_access",
+    "access_specifier",
 )
-class MethodAccess(_FlagHelper, enum.Flag):
+class AccessSpecifier(_FlagHelper, enum.IntFlag):
     public = 0b0000_0001
     private = 0b0000_0010
     abstract = 0b0000_0100
@@ -228,13 +251,6 @@ class Expression:
 @as_tagged_union
 class Literal(Expression):
     """Literal value."""
-    value: Any  # Type specified in subclass
-
-    def __str__(self) -> str:
-        return str(self.value)
-
-
-Constant = Literal  # an alias for now
 
 
 @dataclass
@@ -416,22 +432,88 @@ class Duration(Literal):
         return f"TIME#{self.value}"
 
 
+@dataclass
+@_rule_handler("lduration")
+class Lduration(Literal):
+    """Long duration literal value."""
+
+    days: Optional[lark.Token] = None
+    hours: Optional[lark.Token] = None
+    minutes: Optional[lark.Token] = None
+    seconds: Optional[lark.Token] = None
+    milliseconds: Optional[lark.Token] = None
+    microseconds: Optional[lark.Token] = None
+    nanoseconds: Optional[lark.Token] = None
+    negative: bool = False
+    meta: Optional[Meta] = meta_field()
+
+    @staticmethod
+    def from_lark(minus: Optional[lark.Token], interval: lark.Tree) -> Lduration:
+        kwargs = typing.cast(
+            Dict[str, lark.Token],
+            {
+                tree.data: tree.children[0]
+                for tree in interval.iter_subtrees()
+            }
+        )
+        return Lduration(**kwargs, negative=minus is not None, meta=None)
+
+    @property
+    def value(self) -> str:
+        """The long duration value."""
+        prefix = "-" if self.negative else ""
+        return prefix + "".join(
+            f"{value}{suffix}"
+            for value, suffix in (
+                (self.days, "D"),
+                (self.hours, "H"),
+                (self.minutes, "M"),
+                (self.seconds, "S"),
+                (self.milliseconds, "MS"),
+                (self.microseconds, "US"),
+                (self.nanoseconds, "NS"),
+            )
+            if value is not None
+        )
+
+    def __str__(self):
+        return f"LTIME#{self.value}"
+
+
 @dataclass
 @_rule_handler("time_of_day")
 class TimeOfDay(Literal):
     """Time of day literal value."""
     hour: lark.Token
     minute: lark.Token
+    second: Optional[lark.Token] = None
+    meta: Optional[Meta] = meta_field()
+
+    @property
+    def value(self) -> str:
+        """The time of day value."""
+        return join_if(f"{self.hour}:{self.minute}", ":", self.second)
+
+    def __str__(self):
+        return f"TIME_OF_DAY#{self.value}"
+
+
+@dataclass
+@_rule_handler("ltime_of_day")
+class LtimeOfDay(Literal):
+    """Long time of day literal value."""
+    hour: lark.Token
+    minute: lark.Token
     second: lark.Token
     meta: Optional[Meta] = meta_field()
 
     @property
     def value(self) -> str:
-        """The time of day value."""
+        """The long time of day value."""
         return f"{self.hour}:{self.minute}:{self.second}"
 
     def __str__(self):
-        return f"TIME_OF_DAY#{self.value}"
+        return f"LTIME_OF_DAY#{self.value}"
 
 
 @dataclass
@@ -453,6 +535,25 @@ class Date(Literal):
         return f"DATE#{self.value}"
 
 
+@dataclass
+@_rule_handler("ldate")
+class Ldate(Literal):
+    """Long date literal value."""
+
+    year: lark.Token
+    month: lark.Token
+    day: lark.Token
+    meta: Optional[Meta] = meta_field()
+
+    @property
+    def value(self) -> str:
+        """The long time of day value."""
+        return f"{self.year}-{self.month}-{self.day}"
+
+    def __str__(self):
+        return f"LDATE#{self.value}"
+
+
 @dataclass
 @_rule_handler("date_and_time")
 class DateTime(Literal):
@@ -485,6 +586,38 @@ class DateTime(Literal):
         return f"DT#{self.value}"
 
 
+@dataclass
+@_rule_handler("ldate_and_time")
+class LdateTime(Literal):
+    """Long date and time literal value."""
+
+    ldate: Ldate
+    ltime: LtimeOfDay
+    meta: Optional[Meta] = meta_field()
+
+    @staticmethod
+    def from_lark(
+        year: lark.Token,
+        month: lark.Token,
+        day: lark.Token,
+        hour: lark.Token,
+        minute: lark.Token,
+        second: lark.Token,
+    ) -> LdateTime:
+        return LdateTime(
+            ldate=Ldate(year=year, month=month, day=day),
+            ltime=LtimeOfDay(hour=hour, minute=minute, second=second),
+        )
+
+    @property
+    def value(self) -> str:
+        """The time of day value."""
+        return f"{self.ldate.value}-{self.ltime.value}"
+
+    def __str__(self):
+        return f"LDT#{self.value}"
+
+
 @dataclass
 @_rule_handler("string_literal")
 class String(Literal):
@@ -492,36 +625,53 @@ class String(Literal):
     value: lark.Token
     meta: Optional[Meta] = meta_field()
 
+    def __str__(self) -> str:
+        return str(self.value)
+
 
 @as_tagged_union
 class Variable(Expression):
     ...
 
 
+@dataclass
 @_rule_handler(
     "indirection_type",
     "pointer_type",
 )
-class IndirectionType(Enum):
+class IndirectionType:
     """Indirect access through a pointer or reference."""
-    none = enum.auto()
-    pointer = enum.auto()
-    reference = enum.auto()
+    pointer_depth: int
+    reference: bool
+    meta: Optional[Meta] = meta_field()
+
+    @property
+    def is_indirect(self) -> bool:
+        """True if this denotes a pointer (of any depth) or a reference."""
+        return self.reference or (self.pointer_depth > 0)
 
     @staticmethod
-    def from_lark(token: Optional[lark.Token]) -> IndirectionType:
-        return {
-            "NONE": IndirectionType.none,
-            "POINTER TO": IndirectionType.pointer,
-            "REFERENCE TO": IndirectionType.reference,
-        }[str(token).upper()]
+    def from_lark(*tokens: lark.Token) -> IndirectionType:
+        pointer_depth = 0
+        reference = False
+        upper_tokens = list(map(lambda s: str(s).upper(), tokens))
+        if len(tokens) > 0:
+            pointer_depth = upper_tokens.count("POINTER TO")
+            reference = "REFERENCE TO" in upper_tokens
+        return IndirectionType(
+            pointer_depth=pointer_depth,
+            reference=reference
+        )
+
+    @property
+    def value(self) -> str:
+        return " ".join(
+            ["REFERENCE TO"] * self.reference +
+            ["POINTER TO"] * self.pointer_depth
+        )
 
     def __str__(self):
-        return {
-            IndirectionType.none: "",
-            IndirectionType.pointer: "POINTER TO",
-            IndirectionType.reference: "REFERENCE TO",
-        }[self]
+        return self.value
 
 
 @_rule_handler("incomplete_location")
@@ -800,15 +950,7 @@ class StringTypeInitialization:
     ) -> StringTypeInitialization:
         string_type, length, *value_parts = args
         spec = StringTypeSpecification(string_type, length)
-
-        value: Optional[lark.Token]
-        if len(value_parts):
-            if len(value_parts) == 1:  # lark 0.12.0
-                value, = value_parts
-            else:  # lark 1.0
-                _, value = value_parts
-        else:
-            value = None
+        _, value = value_parts or [None, None]
         return StringTypeInitialization(spec=spec, value=value)
 
     def __str__(self) -> str:
@@ -988,7 +1130,7 @@ class DataType:
     meta: Optional[Meta] = meta_field()
 
     def __str__(self) -> str:
-        if self.indirection and self.indirection != IndirectionType.none:
+        if self.indirection and self.indirection.is_indirect:
             return f"{self.indirection} {self.type_name}"
         return f"{self.type_name}"
 
@@ -1017,7 +1159,7 @@ class IndirectSimpleSpecification:
 @dataclass
 @_rule_handler("array_specification")
 class ArraySpecification:
-    type: DataType
+    type: Union[DataType, FunctionCall]
     subranges: List[Subrange]
     meta: Optional[Meta] = meta_field()
 
@@ -1035,10 +1177,12 @@ class ArraySpecification:
     def from_lark(*args):
         *subranges, type = args
         if isinstance(type, lark.Token):
+            # STRING_TYPE
             type = DataType(
                 indirection=None,
                 type_name=type,
             )
+
         return ArraySpecification(type=type, subranges=subranges)
 
     def __str__(self) -> str:
@@ -1046,38 +1190,37 @@ class ArraySpecification:
         return f"ARRAY [{subranges}] OF {self.type}"
 
 
-ArrayInitialElementType = Union[
-    Constant,
-    "StructureInitialization",
-    EnumeratedValue,
-]
-
-
 @dataclass
 @_rule_handler("array_initial_element")
 class ArrayInitialElement:
     element: ArrayInitialElementType
+    count: Optional[Union[EnumeratedValue, Integer]] = None
     meta: Optional[Meta] = meta_field()
 
     def __str__(self) -> str:
-        return f"{self.element}"
-
-
-@dataclass
-@_rule_handler("array_initial_element_count")
-class ArrayInitialElementCount:
-    count: Union[EnumeratedValue, Integer]
-    element: ArrayInitialElementType
-    meta: Optional[Meta] = meta_field()
-
-    def __str__(self) -> str:
+        if self.count is None:
+            return f"{self.element}"
         return f"{self.count}({self.element})"
 
 
+@_rule_handler("array_initial_element_count")
+class _ArrayInitialElementCount:
+    @staticmethod
+    def from_lark(
+        count: Union[EnumeratedValue, Integer],
+        element: ArrayInitialElementType
+    ) -> ArrayInitialElement:
+        return ArrayInitialElement(
+            element=element,
+            count=count,
+        )
+
+
 @dataclass
 @_rule_handler("array_initialization")
 class ArrayInitialization:
-    elements: List[Union[ArrayInitialElement, ArrayInitialElementCount]]
+    elements: List[ArrayInitialElement]
+    count: Optional[Union[EnumeratedValue, Integer]] = None
     meta: Optional[Meta] = meta_field()
 
     @staticmethod
@@ -1183,6 +1326,7 @@ class StructureElementDeclaration:
         SubrangeTypeInitialization,
         EnumeratedTypeInitialization,
         InitializedStructure,
+        FunctionCall,
     ]
     meta: Optional[Meta] = meta_field()
 
@@ -1201,6 +1345,7 @@ UnionElementSpecification = Union[
     SimpleSpecification,
     SubrangeSpecification,
     EnumeratedSpecification,
+    IndirectSimpleSpecification,
 ]
 
 
@@ -1324,18 +1469,6 @@ class StructureElementInitialization:
         return f"{self.value}"
 
 
-@dataclass
-@_rule_handler("initialized_structure_type_declaration", comments=True)
-class InitializedStructureTypeDeclaration:
-    name: lark.Token
-    extends: Optional[lark.Token]
-    init: StructureInitialization
-    meta: Optional[Meta] = meta_field()
-
-    def __str__(self) -> str:
-        return f"{self.name} : {self.init}"
-
-
 @dataclass
 @_rule_handler("unary_expression")
 class UnaryOperation(Expression):
@@ -1344,13 +1477,8 @@ class UnaryOperation(Expression):
     meta: Optional[Meta] = meta_field()
 
     @staticmethod
-    def from_lark(*args):
-        if len(args) == 1:
-            constant, = args
-            return constant
-
-        operator, expr = args
-        if not operator:
+    def from_lark(operator: Optional[lark.Token], expr: Expression):
+        if operator is None:
             return expr
         return UnaryOperation(
             op=operator,
@@ -1368,6 +1496,8 @@ class UnaryOperation(Expression):
     "expression",
     "add_expression",
     "and_expression",
+    "and_then_expression",
+    "or_else_expression",
     "assignment_expression",
     "xor_expression",
     "comparison_expression",
@@ -1428,11 +1558,40 @@ class FunctionCall(Expression):
     parameters: List[ParameterAssignment]
     meta: Optional[Meta] = meta_field()
 
+    @property
+    def base_type_name(self) -> str:
+        """
+        The base type name.
+
+        This is used as part of the summary mechanism. The "type" is that
+        of the underlying function block or function.
+        """
+        if isinstance(self.name, SimpleVariable):
+            return str(self.name.name)
+        return str(self.name)
+
+    @property
+    def full_type_name(self) -> str:
+        """The full type name, including any dereferencing or subscripts."""
+        return str(self.name)
+
+    @property
+    def value(self) -> str:
+        """
+        The initialization value (the function call itself).
+
+        This is used as part of the summary tool.
+        """
+        return str(self)
+
     @staticmethod
     def from_lark(
         name: SymbolicVariable,
         *parameters: ParameterAssignment,
     ) -> FunctionCall:
+        # Condition parameters (which may be `None`) to represent empty tuple
+        if parameters == (None, ):
+            parameters = []
         return FunctionCall(
             name=name,
             parameters=list(parameters)
@@ -1594,7 +1753,7 @@ class FunctionBlockNameDeclaration(FunctionBlockDeclaration):
 @_rule_handler("fb_invocation_decl", comments=True)
 class FunctionBlockInvocationDeclaration(FunctionBlockDeclaration):
     variables: List[lark.Token]
-    init: FunctionBlockInvocation
+    init: FunctionCall
     meta: Optional[Meta] = meta_field()
 
     def __str__(self) -> str:
@@ -1650,43 +1809,6 @@ class OutputParameterAssignment(ParameterAssignment):
         return prefix + join_if(self.name, " => ", self.value)
 
 
-@dataclass
-@_rule_handler("fb_invocation")
-class FunctionBlockInvocation:
-    name: SymbolicVariable
-    parameters: List[ParameterAssignment]
-    meta: Optional[Meta] = meta_field()
-
-    @property
-    def base_type_name(self) -> lark.Token:
-        """The base type name."""
-        return self.name.name
-
-    @property
-    def full_type_name(self) -> str:
-        """The full type name."""
-        return str(self.name)
-
-    @property
-    def value(self) -> str:
-        """The initialization value (call)."""
-        return str(self)
-
-    @staticmethod
-    def from_lark(
-        name: SymbolicVariable,
-        *parameters: ParameterAssignment,
-    ) -> FunctionBlockInvocation:
-        return FunctionBlockInvocation(
-            name=name,
-            parameters=list(parameters)
-        )
-
-    def __str__(self) -> str:
-        parameters = ", ".join(str(param) for param in self.parameters)
-        return f"{self.name}({parameters})"
-
-
 AnyLocation = Union[Location, IncompleteLocation]
 
 
@@ -1731,7 +1853,7 @@ LocatedVariableSpecInit = Union[
 @_rule_handler("global_var_decl", comments=True)
 class GlobalVariableDeclaration:
     spec: GlobalVariableSpec
-    init: Union[LocatedVariableSpecInit, FunctionBlockInvocation]
+    init: Union[LocatedVariableSpecInit, FunctionCall]
     meta: Optional[Meta] = meta_field()
 
     @property
@@ -1788,7 +1910,7 @@ class Implements:
 @_rule_handler("function_block_type_declaration", comments=True)
 class FunctionBlock:
     name: lark.Token
-    abstract: bool
+    access: Optional[AccessSpecifier]
     extends: Optional[Extends]
     implements: Optional[Implements]
     declarations: List[VariableDeclarationBlock]
@@ -1798,7 +1920,7 @@ class FunctionBlock:
     @staticmethod
     def from_lark(
         fb_token: lark.Token,
-        abstract: Optional[lark.Token],
+        access: Optional[AccessSpecifier],
         derived_name: lark.Token,
         extends: Optional[Extends],
         implements: Optional[Implements],
@@ -1807,7 +1929,7 @@ class FunctionBlock:
         *declarations, body, _ = args
         return FunctionBlock(
             name=derived_name,
-            abstract=abstract is not None,
+            access=access,
             extends=extends,
             implements=implements,
             declarations=list(declarations),
@@ -1815,8 +1937,8 @@ class FunctionBlock:
         )
 
     def __str__(self) -> str:
-        abstract = "ABSTRACT " if self.abstract else ""
-        header = f"FUNCTION_BLOCK {abstract}{self.name}"
+        access_and_name = join_if(self.access, " ", self.name)
+        header = f"FUNCTION_BLOCK {access_and_name}"
         header = join_if(header, " ", self.implements)
         header = join_if(header, " ", self.extends)
         return "\n".join(
@@ -1834,6 +1956,7 @@ class FunctionBlock:
 @dataclass
 @_rule_handler("function_declaration", comments=True)
 class Function:
+    access: Optional[AccessSpecifier]
     name: lark.Token
     return_type: Optional[Union[SimpleSpecification, IndirectSimpleSpecification]]
     declarations: List[VariableDeclarationBlock]
@@ -1842,6 +1965,7 @@ class Function:
 
     @staticmethod
     def from_lark(
+        access: Optional[AccessSpecifier],
         name: lark.Token,
         return_type: Optional[Union[SimpleSpecification, IndirectSimpleSpecification]],
         *remainder
@@ -1849,6 +1973,7 @@ class Function:
         *declarations, body = remainder
         return Function(
             name=name,
+            access=access,
             return_type=return_type,
             declarations=typing.cast(
                 List[VariableDeclarationBlock], list(declarations)
@@ -1857,7 +1982,8 @@ class Function:
         )
 
     def __str__(self) -> str:
-        function = f"FUNCTION {self.name}"
+        access_and_name = join_if(self.access, " ", self.name)
+        function = f"FUNCTION {access_and_name}"
         return_type = f": {self.return_type}" if self.return_type else None
         return "\n".join(
             line for line in
@@ -1913,7 +2039,7 @@ class Action:
 @dataclass
 @_rule_handler("function_block_method_declaration", comments=True)
 class Method:
-    access: Optional[MethodAccess]
+    access: Optional[AccessSpecifier]
     name: lark.Token
     return_type: Optional[LocatedVariableSpecInit]
     declarations: List[VariableDeclarationBlock]
@@ -1922,7 +2048,7 @@ class Method:
 
     @staticmethod
     def from_lark(
-        access: Optional[MethodAccess],
+        access: Optional[AccessSpecifier],
         name: lark.Token,
         return_type: Optional[LocatedVariableSpecInit],
         *args
@@ -1954,7 +2080,7 @@ class Method:
 @dataclass
 @_rule_handler("function_block_property_declaration", comments=True)
 class Property:
-    access: Optional[MethodAccess]
+    access: Optional[AccessSpecifier]
     name: lark.Token
     return_type: Optional[LocatedVariableSpecInit]
     declarations: List[VariableDeclarationBlock]
@@ -1963,7 +2089,7 @@ class Property:
 
     @staticmethod
     def from_lark(
-        access: Optional[MethodAccess],
+        access: Optional[AccessSpecifier],
         name: lark.Token,
         return_type: Optional[LocatedVariableSpecInit],
         *args
@@ -2020,6 +2146,12 @@ class VariableDeclarationBlock:
     items: List[Any]
     meta: Optional[Meta]
 
+    @property
+    def attribute_pragmas(self) -> List[str]:
+        """Attribute pragmas associated with the variable declaration block."""
+        # TODO: deprecate
+        return getattr(self.meta, "attribute_pragmas", [])
+
 
 @dataclass
 @_rule_handler("var_declarations", comments=True)
@@ -2056,8 +2188,10 @@ class StaticDeclarations(VariableDeclarationBlock):
 
     @staticmethod
     def from_lark(
-        attrs: Optional[VariableAttributes], *items: VariableInitDeclaration
+        attrs: Optional[VariableAttributes],
+        tree: lark.Tree,
     ) -> StaticDeclarations:
+        items = typing.cast(List[VariableInitDeclaration], tree.children)
         return StaticDeclarations(attrs, list(items))
 
     def __str__(self) -> str:
@@ -2409,14 +2543,14 @@ class AccessDeclarations(VariableDeclarationBlock):
 @_rule_handler("global_var_declarations", comments=True)
 class GlobalVariableDeclarations(VariableDeclarationBlock):
     block_header: ClassVar[str] = "VAR_GLOBAL"
-    attrs: Optional[VariableAttributes]
+    attrs: Optional[GlobalVariableAttributes]
     items: List[GlobalVariableDeclaration]
     meta: Optional[Meta] = meta_field()
     name: Optional[str] = None
 
     @staticmethod
     def from_lark(
-        attrs: Optional[VariableAttributes],
+        attrs: Optional[GlobalVariableAttributes],
         *items: GlobalVariableDeclaration
     ) -> GlobalVariableDeclarations:
         return GlobalVariableDeclarations(
@@ -2425,20 +2559,6 @@ class GlobalVariableDeclarations(VariableDeclarationBlock):
             items=list(items)
         )
 
-    @property
-    def attribute_pragmas(self) -> Set[str]:
-        """Attribute pragmas."""
-        if self.meta is None:
-            return set()
-
-        _, pragmas = self.meta.get_comments_and_pragmas()
-        attributes = set()
-        for pragma in pragmas:
-            # TODO: better pragma parsing; it's its own grammar
-            if pragma.startswith("{attribute "):  # }
-                attributes.add(pragma.split(" ")[1].strip(" }'"))
-        return attributes
-
     def __str__(self) -> str:
         return "\n".join(
             (
@@ -2454,13 +2574,13 @@ class Statement:
     ...
 
 
-@_rule_handler("fb_invocation_statement", comments=True)
-class FunctionBlockInvocationStatement(Statement, FunctionBlockInvocation):
+@_rule_handler("function_call_statement", comments=True)
+class FunctionCallStatement(Statement, FunctionCall):
     @staticmethod
     def from_lark(
-        invocation: FunctionBlockInvocation,
-    ) -> FunctionBlockInvocationStatement:
-        return FunctionBlockInvocationStatement(
+        invocation: FunctionCall,
+    ) -> FunctionCallStatement:
+        return FunctionCallStatement(
             name=invocation.name,
             parameters=invocation.parameters,
             meta=invocation.meta,
@@ -2545,7 +2665,14 @@ class IfStatement(Statement):
         )
 
 
-CaseMatch = Union[Subrange, Integer, EnumeratedValue, SymbolicVariable]
+CaseMatch = Union[
+    Subrange,
+    Integer,
+    EnumeratedValue,
+    SymbolicVariable,
+    BitString,
+    Boolean,
+]
 
 
 @dataclass
@@ -2772,16 +2899,7 @@ class StatementList:
         )
 
     def __str__(self) -> str:
-        def stringify_statement(statement: Union[Statement, FunctionBlockInvocation]) -> str:
-            # TODO: this is a bit of a bug; who has the responsibility to
-            # append a semicolon?
-            if not isinstance(statement, Statement):
-                return f"{statement};"
-            return str(statement)
-
-        return "\n".join(
-            stringify_statement(statement) for statement in self.statements
-        )
+        return "\n".join(str(statement) for statement in self.statements)
 
 
 FunctionBlockBody = Union[
@@ -2806,13 +2924,15 @@ TypeDeclarationItem = Union[
 @_rule_handler("data_type_declaration", comments=True)
 class DataTypeDeclaration:
     declaration: Optional[TypeDeclarationItem]
+    access: Optional[AccessSpecifier]
     meta: Optional[Meta] = meta_field()
 
     @staticmethod
     def from_lark(
+        access: Optional[AccessSpecifier],
         declaration: Optional[TypeDeclarationItem] = None,
     ) -> DataTypeDeclaration:
-        return DataTypeDeclaration(declaration)
+        return DataTypeDeclaration(access=access, declaration=declaration)
 
     def __str__(self) -> str:
         if not self.declaration:
@@ -2825,6 +2945,8 @@ class DataTypeDeclaration:
             # note: END_STRUCT; END_UNION; result in "END_TYPE expected not ;"
             decl = decl + ";"
 
+        decl = join_if(self.access, " ", decl)
+
         return "\n".join(
             (
                 f"TYPE {decl}",
@@ -2900,7 +3022,7 @@ class GrammarTransformer(lark.visitors.Transformer_InPlaceRecursive):
     def __init__(
         self,
         comments: Optional[List[lark.Token]] = None,
-        fn: Optional[str] = None,
+        fn: Optional[AnyPath] = None,
         source_code: Optional[str] = None,
     ):
         super().__init__()
@@ -3031,13 +3153,40 @@ def merge_comments(source: Any, comments: List[lark.Token]):
                 merge_comments(obj, comments)
 
 
+Constant = Union[
+    Duration,
+    Lduration,
+    TimeOfDay,
+    Date,
+    DateTime,
+    Ldate,
+    LdateTime,
+    Real,
+    Integer,
+    String,
+    BitString,
+    Boolean,
+]
+
+
+ArrayInitialElementType = Union[
+    Constant,
+    StructureInitialization,
+    EnumeratedValue,
+]
+
+
 if apischema is not None:
     # Optional apischema deserializers
 
     @apischema.deserializer
-    def _method_access_deserializer(access: int) -> MethodAccess:
-        return MethodAccess(access)
+    def _method_access_deserializer(access: int) -> AccessSpecifier:
+        return AccessSpecifier(access)
 
     @apischema.deserializer
     def _var_attrs_deserializer(attrs: int) -> VariableAttributes:
         return VariableAttributes(attrs)
+
+    @apischema.deserializer
+    def _global_var_attrs_deserializer(attrs: int) -> GlobalVariableAttributes:
+        return GlobalVariableAttributes(attrs)
